import timeit
start = timeit.default_timer()

from ruamel import yaml
import VoyageGenerator_heur1 as VG
import TravelRouteDataConfiguration as TRDC
import TotalResourceCapacity as TRC

inputFilepath = 'test_instances/R_nSR_8_nTasks_30_nLoc_8_1.txt'
outputFilepath = inputFilepath[:-4] + '_TravelRoutes.txt'
print(outputFilepath)

#travelRouteVariables er en liste, der hvert element representerer en supperressurs og er en liste der
#   hvert element representerer en rute, og er en liste der
#   hvert element representerer en y, og er en liste der
#   hvert element er en indeks, i rekkefølger g, i, m, j, n
#   For å accesse en y brukes 0-indeksering, men den lagra informasjonen er 1-indeksert
#   Første y-variabel for andre rute til tredje super-ressurs hentes slik: y = travelRouteVariables[2][1][0]
#   og ser slik ut: y = [g, i, m, j, n]
def labelToVariables(superResourceNr, labels):
    variables = []
    for label in labels:
        labelVariables = []
        for i in range(0, len(label.route)-1):
            nodeFrom = label.route[i]
            nodeTo = label.route[i+1]
            labelVariables.append([superResourceNr, nodeFrom, label.route[:i+1].count(nodeFrom), nodeTo, label.route[:i+2].count(nodeTo)])
        variables.append(labelVariables)
    return variables

def variableGenerator():
    with open("test_gen_inputdata.yaml", 'r') as file:
        read_data = yaml.load(file, Loader=yaml.Loader)

        global nSR
        nSR = read_data['nSR']
        global nDuplicates
        nDuplicates = read_data['nDuplicates']
        global nLoc
        nLoc = read_data['nLoc']
        global nSkillsLow
        nSkillsLow = read_data['nSkillsLow']
        global nSkillsMedium
        nSkillsMedium= read_data['nSkillsHigh']
        global nSkillsHigh
        nSkillsHigh = read_data['nSkillsHigh']
        global capreq_min
        capreq_min= read_data['capreq_min']
        global capreq_max
        capreq_max = read_data['capreq_max']
        global nSkillsMin
        nSkillsMin = read_data['nSkillsMin']
        global nSkillsMax
        nSkillsMax = read_data['nSkillsMax']
        global capPerSkillMin
        capPerSkillMin = read_data['capPerSkillMin']
        global capPerSkillMax
        capPerSkillMax = read_data['capPerSkillMax']
        global routeSelection
        routeSelection = read_data['routeSelection']
        global nResourcesPerTask
        nResourcesPerTask = read_data['nResourcesPerTask']

    #Lage dict over alle typer
    parametersSingle = {"nLocations:": 0, "nEndTime:": 0, "nSkills:": 0, "nTasks:": 0, "nSuperResources:": 0, "nSubResources:": 0}
    parametersVector = {"Deadline:": [], "ReleaseTime:": [], "Tmin:": [], "Ttask:": [], "BelongToTaskLong:": [], "BelongToTaskSleep:": [], "BelongToTaskRest:" : [], "BelongToTaskDuplicate:": [], "BelongToLocation:": []}
    parametersMatrix = {"Ttravel:": [], "TaskSkills:": [], "CapacityRequirement:": [], "Precedence: [": [], "Connected: [": [], "DirectStart: [": [],  "Value: [": []}

    #Finne alle parametere en etter en og legge inn på dict
    TRDC.getSingleVariables(inputFilepath, parametersSingle)
    TRDC.getVectorVariables(inputFilepath, parametersVector)
    TRDC.getMatrixVariables(inputFilepath, parametersMatrix)

    #Assign variables
    nLocations = parametersSingle["nLocations:"]
    nEndTime = parametersSingle["nEndTime:"]
    nSkills = parametersSingle["nSkills:"]
    nTasks = parametersSingle["nTasks:"]
    nSuperResources = parametersSingle["nSuperResources:"]
    nSubResources = parametersSingle["nSubResources:"]

    Deadline = parametersVector["Deadline:"]
    ReleaseTime = parametersVector["ReleaseTime:"]
    Tmin = parametersVector["Tmin:"]
    Ttask = parametersVector["Ttask:"]
    BelongToTaskLong = parametersVector["BelongToTaskLong:"]
    BelongToTaskSleep = parametersVector["BelongToTaskSleep:"]
    BelongToTaskRest = parametersVector["BelongToTaskRest:"]
    BelongToTaskDuplicate = parametersVector["BelongToTaskDuplicate:"]
    BelongToLocation = parametersVector["BelongToLocation:"]


    Ttravel = parametersMatrix["Ttravel:"] #Default er en liste av matriser
    TaskSkills = parametersMatrix["TaskSkills:"][0] #For alle matriser som bare er en matrise, må den hentes ut
    CapacityRequirement = parametersMatrix["CapacityRequirement:"][0]
    Precedence = parametersMatrix["Precedence: ["][0]
    Connected = parametersMatrix["Connected: ["][0]
    DirectStart = parametersMatrix["DirectStart: ["][0]
    Value = parametersMatrix["Value: ["][0]

    superResources = TRDC.getSuperResourceDict(nSuperResources)
    skillDict = TRDC.getSkillDict(nSubResources)
    tasksAtLocation = TRDC.getTasksAtLocation(nLocations, BelongToLocation)
    solvableTasksList = VG.genSolvableTasksList(nTasks, nSkills, nDuplicates, BelongToTaskDuplicate, superResources,
                                                skillDict, TaskSkills, Precedence, Connected, DirectStart,
                                                nResourcesPerTask, CapacityRequirement)
    securityResource = 0
    lowestCapacity = nSkills * nSubResources
    lowestResourceNumberofTasks = len(solvableTasksList) + 1
    travelRouteVariables = []



    for superResource, subResources in superResources.items():
        superResourceSkills = VG.generateSuperResourceSkillList(subResources, skillDict, nSkills)
        resourceNumberofTasks = 0
        SuperResourceCapacity = TRC.addResourceCapacity(subResources, skillDict, nSkills)
        for task in solvableTasksList:
            for skill in range(1, nSkills-1):
                if superResourceSkills[skill] and TaskSkills[task-1][skill]:
                    resourceNumberofTasks += 1
                    break
        if resourceNumberofTasks < lowestResourceNumberofTasks:
            lowestResourceNumberofTasks = resourceNumberofTasks
            securityResource = superResource
            lowestCapacity = SuperResourceCapacity
        if resourceNumberofTasks == lowestResourceNumberofTasks:
            if SuperResourceCapacity <= lowestCapacity:
                lowestResourceNumberofTasks = resourceNumberofTasks
                securityResource = superResource
                lowestCapacity = SuperResourceCapacity
    for superResource, subResources in superResources.items():
        finalLabels = VG.generateLabels(nLocations, nEndTime, superResource, subResources, skillDict, nSkills,
                                        tasksAtLocation, TaskSkills, Deadline, ReleaseTime, Tmin, Ttask, Ttravel,
                                        BelongToTaskLong, BelongToTaskSleep, BelongToTaskDuplicate, solvableTasksList,
                                        securityResource, Value, CapacityRequirement, routeSelection)
        travelRouteVariables.append(labelToVariables(superResource, finalLabels))

        print("Finished with superresource", superResource, ", number of labels:", len(finalLabels))
    return travelRouteVariables, securityResource


def saveVariables(travelRouteVariables, securityResource):
    nTravelRoutesOfG = [0 for _ in range(0, len(travelRouteVariables))]
    YlocationRoutes = []
    i = 0
    with open(outputFilepath, "w") as file:
        outputFile = outputFilepath.split('/')[-1]
        file.write('!' + outputFile + '\n')
        spaces = " " * 100 * 10
        file.write(spaces + '\n')
        file.write(spaces + '\n')
        file.write(spaces + '\n')
        file.write(spaces + '\n')
        file.write('Security: ' + str(securityResource) + '\n\n')
        file.write('YtravelRoutes: [')
        for superResourceVariables in travelRouteVariables:
            for route in superResourceVariables:
                nTravelRoutesOfG[i] += 1
                for variable in route:
                    YlocationRoutes.append([nTravelRoutesOfG[i], variable[0], variable[1], variable[2]])
                    file.write(
                        '(' + str(nTravelRoutesOfG[i]) + ' ' + str(variable[0]) + ' ' + str(variable[1]) + ' ' + str(
                            variable[2]) + ' ' + str(variable[3]) + ' ' + str(variable[4]) + ') 1 ')
                YlocationRoutes.append([nTravelRoutesOfG[i], i + 1, route[-1][3], 1])
            i += 1
        file.write(']\n')

        file.write('YlocationRoutes: [')
        for data in YlocationRoutes:
            file.write('(' + str(data[0]) + ' ' + str(data[1]) + ' ' + str(data[2]) + ' ' + str(data[3]) + ') 1 ')
        file.write(']\n')

        file.seek(len(outputFile) + 2)
        file.write('\nnTravelRoutesOfG: [')
        for i in nTravelRoutesOfG:
            file.write(str(i) + ' ')
        file.write(']\n')

        sumtravelroutes = sum(nTravelRoutesOfG)

        file.write('nRoutesMax: ' + str(max(nTravelRoutesOfG)) + '\n')
        stop = timeit.default_timer()
        runtime = stop - start
        file.write('!Runtime: ' + str(runtime) + '\n')
        file.write('!SumTravelRoutes: ' + str(sumtravelroutes) + '\n')
        print(runtime)


travelRouteVariables, securityResource = variableGenerator()
saveVariables(travelRouteVariables, securityResource)